# Copyright 2005 David LaPorte <david@davidlaporte.org>
# Copyright 2005 Kevin Amorin <kev@amorin.org>
#
# See the enclosed file COPYING for license information (GPL).
# If you did not receive this file, see
# http://www.fsf.org/licensing/licenses/gpl.html.
#

package pf::trigger;

use strict;
use warnings;

our ($trigger_desc_sql, $trigger_view_vid_sql, $trigger_view_sql, $trigger_view_enable_sql,
     $trigger_view_all_sql, $trigger_exist_sql, $trigger_view_type_sql, $trigger_add_sql,
     $trigger_delete_vid_sql, $trigger_delete_all_sql, $trigger_view_tid_sql);

BEGIN {
  use Exporter ();
  our (@ISA, @EXPORT);
  @ISA    = qw(Exporter);
  @EXPORT = qw(trigger_db_prepare trigger_view trigger_view_enable trigger_view_all trigger_delete_all 
               trigger_in_range trigger_add trigger_view_type trigger_view_tid trigger_scan);
}

use lib qw(/usr/local/pf/lib);
use pf::config;
use pf::db;
use pf::util;
use pf::violation qw(violation_trigger);
use pf::iplog qw(ip2mac);

trigger_db_prepare($dbh) if (!$thread);

sub trigger_db_prepare {
  my ($dbh) = @_;
  $trigger_desc_sql=$dbh->prepare( qq [ desc trigger ] );
  $trigger_view_sql=$dbh->prepare( qq[ select tid_start,tid_end,class.vid,type,description from trigger,class where class.vid=trigger.vid and tid_start<=? and tid_end>=? and type=?]);
  $trigger_view_enable_sql=$dbh->prepare( qq[ select tid_start,tid_end,class.vid,type from trigger,class where class.vid=trigger.vid and tid_start<=? and tid_end>=? and type=? and disable="N"]);
  $trigger_view_vid_sql=$dbh->prepare( qq[ select tid_start,tid_end,class.vid,description from trigger,class where class.vid=trigger.vid and class.vid=?]);
  $trigger_view_tid_sql=$dbh->prepare( qq[ select tid_start,tid_end,class.vid,type,description from trigger,class where class.vid=trigger.vid and tid_start<=? and tid_end>=? ]);
  $trigger_view_all_sql=$dbh->prepare( qq[ select tid_start,tid_end,class.vid,type,description from trigger,class where class.vid=trigger.vid ]);
  $trigger_view_type_sql=$dbh->prepare( qq[ select tid_start,tid_end,class.vid,type,description from trigger,class where class.vid=trigger.vid and type=?]);
  $trigger_exist_sql=$dbh->prepare( qq [ select vid,tid_start,tid_end,type from trigger where vid=? and tid_start<=? and tid_end>=? and type=?]);
  $trigger_add_sql=$dbh->prepare( qq [ insert into trigger(vid,tid_start,tid_end,type) values(?,?,?,?) ]);
  $trigger_delete_vid_sql=$dbh->prepare( qq [ delete from trigger where vid=? ]);
  $trigger_delete_all_sql=$dbh->prepare( qq [ delete from trigger ]);
}

sub trigger_desc {
 return db_data($trigger_desc_sql);
}

sub trigger_view {
  my ($tid,%type) = @_;
  return db_data($trigger_view_sql,$tid,$tid,$type{type});
}

sub trigger_view_enable {
  my ($tid,$type) = @_;
  return db_data($trigger_view_enable_sql,$tid,$tid,$type);
}

sub trigger_view_vid {
  my ($vid) = @_;
  return db_data($trigger_view_vid_sql,$vid);
}  

sub trigger_view_tid {
  my ($tid) = @_;
  return db_data($trigger_view_tid_sql,$tid,$tid);
}

sub trigger_view_all {
  return db_data($trigger_view_all_sql);
}     

sub trigger_view_type {
  my ($type) = @_;
  return db_data($trigger_view_type_sql,$type);
}     

sub trigger_delete_vid {
  my ($vid) = @_;
  $trigger_delete_vid_sql->execute($vid) || return(0);
  pflogger("triggers vid $vid deleted", 2);
  return(1);
}

sub trigger_delete_all {
  $trigger_delete_all_sql->execute() || return(0);
  pflogger("All triggers deleted", 2);
  return(1);
}

sub trigger_exist {
  my ($vid,$tid_start,$tid_end,$type) = @_;
  $trigger_exist_sql->execute($vid,$tid_start,$tid_end,$type) || return(0);
  my ($val) = $trigger_exist_sql->fetchrow_array();
  $trigger_exist_sql->finish();
  return($val);
}

#
# clean input parameters and add to trigger table
#
sub trigger_add {
  my ($vid,$tid_start,$tid_end,$type)=@_;
  if (trigger_exist($vid,$tid_start,$tid_end,$type)) {
    pflogger("attempt to add existing trigger $tid_start $tid_end [$type]", 1);
    return(2);
  }
  $trigger_add_sql->execute($vid,$tid_start,$tid_end,$type) || return(0);
  pflogger("trigger $tid_start $tid_end added", 2);
  return(1);
}

sub trigger_scan {
  my ($addr,$tids)=@_;
  eval "use Net::Nessus::ScanLite; 1" || return(0);
  my @return;
  return -1 if (!$addr);
  my $host = $Config{'scan'}{'host'};
  my $port = $Config{'scan'}{'port'};
  my $user = $Config{'scan'}{'user'};
  my $pass = $Config{'scan'}{'pass'};
  my $ssl = isenabled($Config{'scan'}{'ssl'});
  my @ref = trigger_view_type("scan");
  my $pluginlist;
  my $holefound=0;
  if (!$tids || $tids=~/^all$/i){
    foreach my $row (@ref) {
      for (my $i=$row->{'tid_start'};$i<=$row->{'tid_end'};$i++){
        $pluginlist.="$i;";
      }
    }
    chop($pluginlist);
  }else{
    return -1 if ($tids!~/^[\d|;| ]+$/);
    $pluginlist=$tids;
  }
  return 0 if (!$pluginlist);
  my $trigger = Net::Nessus::ScanLite->new(  host            => $host,
                                            port            => $port,
                                            ssl             => $ssl,
                                          );

  $trigger->preferences( { host_expansion => 'none', safe_checks => 'yes', checks_read_timeout => 1 });
  $trigger->plugin_set($pluginlist);
  &pflogger("plug = $pluginlist addr = $addr user =$user port = $port host = $host ssl = $ssl\n",2);
  if ($trigger->login($user,$pass)) {
        &pflogger("starting to trigger node(s) $addr ", 2);
        $trigger->attack($addr);
        &pflogger("Address $addr total info: ".$trigger->total_info." Total holes: ".$trigger->total_holes,2);
        push @return, "Address $addr total info: ".$trigger->total_info." Total holes: ".$trigger->total_holes."\n";   
        foreach my $info ($trigger->info_list){
                my $srcmac = ip2mac($info->Host);
                push (@return,join $main::delimiter,(0,$srcmac,$info->Host,$info->ScanID, 0, $info->Port,$info->Description));
                pflogger("Info ID: ". $info->ScanID."  Host: ".$info->Host." Port: ".$info->Port,8);
        }
        foreach my $info ($trigger->hole_list) {
                my $tid=$info->ScanID;
                my $srcmac = ip2mac($info->Host);
                push (@return,join $main::delimiter, (1,$srcmac,$info->Host, $tid, $info->Port,$info->Description));
                pflogger("Hole ID: ". $info->ScanID."  Host: ".$info->Host." Port: ".$info->Port,2);
                if (defined $Config{'scan'}{'live_tids'} && grep(/^$tid$/,split(/\s*,\s*/, $Config{'scan'}{'live_tids'}))){
                  pflogger("Trying to add trigger $tid for ($srcmac) (".$info->Host.")", 2);
                  violation_trigger($srcmac, $tid, "scan") if ($srcmac);
                }else {
                  pflogger("NOT ADDING Trigger - $tid for $srcmac (".$info->Host.") please add $tid to scan.livetids if you would like this done",1);
                }
        }
  } else {
        push(@return,"trigger_scan: Nessus login failed: ".$trigger->code.": ".$trigger->error."\n");
        pflogger("Nessus login failed: ".$trigger->code.": ".$trigger->error,1);
  }

   return(@return);
}

sub trigger_in_range {
  my($range, $trigger) = @_;           
  foreach my $element (split(/\s*,\s*/, $range)) {
    if ($element eq $trigger) {
      return(1);
    } elsif ($element =~ /^\d+\s*\-\s*\d+$/) {
      my ($begin, $end) = split(/\s*\-\s*/, $element);
      if ($trigger >= $begin && $trigger <= $end) {
        return(1)
      }    
    } else {
      return(0);
    }            
  }
}

1
