#
	# $Id: util.pm,v 1.6 2005/11/30 21:41:09 kevmcs Exp $
#
# Copyright 2005 David LaPorte <david@davidlaporte.org>
# Copyright 2005 Kevin Amorin <kev@amorin.org>
#
# See the enclosed file COPYING for license information (GPL).
# If you did not receive this file, see
# http://www.fsf.org/licensing/licenses/gpl.html.
#

package pf::util;

use strict;
use warnings;
use File::Basename;
use FileHandle;
use Sys::Syslog;
use POSIX();
use Net::SMTP;

our (%trappable_ip, %reggable_ip, %is_internal);

BEGIN {
  use Exporter ();
  our (@ISA, @EXPORT);
  @ISA    = qw(Exporter);
  @EXPORT = qw(valid_date valid_ip clean_mac valid_mac whitelisted_mac trappable_mac trappable_ip reggable_ip
               inrange_ip ip2gateway ip2interface ip2device isinternal pflogger pfmailer isenabled
               isdisabled getlocalmac ip2int int2ip get_all_internal_ips get_internal_nets get_internal_ips 
               get_internal_devs get_internal_devs_phy get_external_devs get_managed_devs get_internal_macs 
               get_internal_info get_gateways get_dhcp_devs num_interfaces createpid readpid deletepid 
               parse_template mysql_date util_funnyarp oui_to_vendor normalize_time throw_hissy_fit);
}

use lib qw(/usr/local/pf/lib);
use pf::config;

if (basename($0) eq "pfmon" && isenabled($Config{'general'}{'caching'})) {
  %trappable_ip = preload_trappable_ip();
  %reggable_ip  = preload_reggable_ip();
  %is_internal  = preload_is_internal();
}

# pflogger logging levels
# 1 - errors
# 2 - interesting info
# 4 - needed information (on large changes)
# 8 - normal operations
# 10 - database queries
# 12 - extreme arp operations
# 16 - debugginig only
#
sub valid_date {
  my ($date) = @_;
  # kludgy but short
  if ($date !~ /^\d{4}\-((0[1-9])|(1[0-2]))\-((0[1-9])|([12][0-9])|(3[0-1]))\s+(([01][0-9])|(2[0-3]))(:[0-5][0-9]){2}$/) {
    pflogger("invalid date $date",1);
    return(0);
   } else {
    return(1);
  }
}

sub valid_ip {
  my ($ip) = @_;
  if (!$ip || $ip !~ /^(?:\d{1,3}\.){3}\d{1,3}$/ || $ip =~ /^0\.0\.0\.0$/) {
    my $caller = (caller(1))[3] || basename($0);
    $caller =~ s/^(pf::\w+|main):://;
    pflogger("invalid IP: $ip from $caller", 1);
    pflogger("invalid IP: $ip", 1);
    return(0);
  } else {
    return(1);
  }
}

sub clean_mac {
  my ($mac) = @_;
  return(0) if (!$mac);
  $mac =~ s/\s//g;
  $mac = lc($mac);
  $mac =~ s/\.//g if ($mac =~ /^([0-9a-f]{4}(\.|$)){4}$/i);
  $mac =~ s/([a-f0-9]{2})(?!$)/$1:/g if ($mac =~ /^[a-f0-9]{12}$/i);
  $mac = join q {:} => map {sprintf "%02x" => hex} split m {:|\-} => $mac;
  return($mac);
}

sub valid_mac {
  my ($mac) = @_;
  $mac = clean_mac($mac);
  if ($mac =~ /^ff:ff:ff:ff:ff:ff$/ || $mac =~ /^00:00:00:00:00:00$/ || $mac !~ /^([0-9a-f]{2}(:|$)){6}$/i) {
    pflogger("invalid MAC: $mac", 1);
    return(0);
  } else {
    return(1);
  }
}

sub whitelisted_mac {
  my ($mac) = @_;
  return(0) if (!valid_mac($mac));
  $mac = clean_mac($mac);
  foreach my $whitelist (split(/\s*,\s*/, $Config{'trapping'}{'whitelist'})) {
    if ($mac eq clean_mac($whitelist)) {
      pflogger("$mac is whitelisted, skipping", 8);
      return(1);
    }
  }
  return(0);
}

sub trappable_mac {
  my ($mac) = @_;
  return(0) if (!$mac);
  $mac = clean_mac($mac);
  #if (!valid_mac($mac) || whitelisted_mac($mac) || $mac eq getlocalmac(ip2device(mac2ip($mac))) || $mac eq $blackholemac) {
  if (!valid_mac($mac) || whitelisted_mac($mac) || grep(/^$mac$/,get_internal_macs()) || $mac eq $blackholemac ) {
    pflogger("$mac is not trappable, skipping", 8);
    return(0);
  } else {
    return(1);
  }
}

sub trappable_ip {
  my ($ip) = @_;
  return(0) if (!$ip || !valid_ip($ip));
  return($trappable_ip{$ip}) if (defined($trappable_ip{$ip}));
  return inrange_ip($ip,$Config{'trapping'}{'range'});
}

sub reggable_ip {
  my ($ip) = @_;
  return(0) if (!$ip || !valid_ip($ip));
  return(1) if (!defined $Config{'registration'}{'range'} || !$Config{'registration'}{'range'});
  return($reggable_ip{$ip}) if (defined($reggable_ip{$ip}));
  return inrange_ip($ip,$Config{'registration'}{'range'});
} 


sub inrange_ip { 
  my ($ip,$network_range) = @_;

  if (grep(/^$ip$/, get_gateways())) {
    pflogger("$ip is a gateway, skipping", 8);
    return(0);
  }
  if (grep(/^$ip$/, get_internal_ips())) {
    pflogger("$ip is a local int, skipping", 8);
    return(0);
  }

  foreach my $range (split(/\s*,\s*/, $network_range)) {
    if ($range =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\/\d{1,2}$/) {
      my $block = new Net::Netmask($range);
      if ($block->size()>2){
        return(1) if ($block->match($ip) && $block->nth(0) ne $ip && $block->nth(-1) ne $ip);
      }else{
        return(1) if ($block->match($ip));
      }  
      #return(1) if ($block->match($ip) && $block->nth(0) ne $ip && $block->nth(-1) ne $ip);
    } elsif ($range =~ /^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})-(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})$/) {

      my $int_ip = ip2int($ip);
      my $start  = $1;
      my $end    = $2;

      if (!valid_ip($start) || !valid_ip($end)) {
        pflogger("$range not valid range!",1);
      } else {
        my $int_start = ip2int($start);
        my $int_end   = ip2int($end);
        return (1) if ($int_ip >= $int_start && $int_ip <= $int_end); 
      }
    } elsif ($range =~ /^(\d{1,3}\.\d{1,3}\.\d{1,3})\.(\d{1,3})-(\d{1,3})$/) {

      my $int_ip = ip2int($ip);
      my $net    = $1;
      my $start  = $2;
      my $end    = $3;

      if (!valid_ip($net.".".$start) || $end < $start || $end > 255) {
        pflogger("$range not valid range!",1);
      } else {
        my $int_start = ip2int($net.".".$start);
        my $int_end   = ip2int($net.".".$end);
        return (1) if ($int_ip >= $int_start && $int_ip <= $int_end); 
      }
	} elsif ($range =~ /^(?:\d{1,3}\.){3}\d{1,3}$/) {
      return (1) if ($range=~/^$ip$/); 	
	} else {
      pflogger("$range not valid!",1);
      next;
    }
  }
  pflogger("$ip is not in $network_range, skipping", 14);
  return(0);
}

sub ip2gateway {
  my ($ip) = @_;
  return(0) if (!valid_ip($ip));
  foreach my $interface (@internal_nets) {
    if ($interface->match($ip)) {
      return($interface->tag("gw"));
    }
  }
  return(0);
}

sub ip2interface {
  my ($ip) = @_;
  return(0) if (!valid_ip($ip));
  foreach my $interface (@internal_nets) {
    if ($interface->match($ip)) {
      return($interface->tag("ip"));
    }
  }
  return(0);
}

sub ip2device {
  my ($ip) = @_;
  return(0) if (!valid_ip($ip));
  foreach my $interface (@internal_nets) {
    if ($interface->match($ip)) {
      return($interface->tag("int"));
    }
  }
  return(0);
}

sub isinternal {
  my ($ip) = @_;
  return(0) if (!valid_ip($ip));
  return($is_internal{$ip}) if (defined($is_internal{$ip}));
  foreach my $interface (@internal_nets) {
    if ($interface->match($ip)) {
      return(1);
    }
  }
  return(0);
}

sub pflogger {
  my ($msg, $v) = @_;
  $v = 4 if (!$v);
  if ($v <= $verbosity) {
    openlog("pf",'',$facility);

    my $caller = (caller(1))[3] || basename($0);
    $caller =~ s/^(pf::\w+|main):://;

    syslog($priority, "$caller: $msg");
    closelog();
  }
}

sub pfmailer {
  my (%data) = @_;
  my $smtpserver = $Config{'alerting'}{'smtpserver'};
  my @to = split(/\s*,\s*/, $Config{'alerting'}{'emailaddr'});
  my $from = 'root@'.$fqdn;
  my $subject = $data{'subject'};
  my $date = POSIX::strftime("%m/%d/%y %H:%M:%S", localtime);
  my $smtp = Net::SMTP->new($smtpserver, Hello => $fqdn );

  if (defined $smtp){
    $smtp->mail($from);
    $smtp->to(@to);
    $smtp->data();
    $smtp->datasend("From: $from\n");
    $smtp->datasend("To: ".join(",",@to)."\n");
    $smtp->datasend("Subject: PF Alert: $subject ($date)\n");
    $smtp->datasend("\n");
    $smtp->datasend($data{'message'});
    $smtp->dataend();
    $smtp->quit;
    pflogger("email regarding '$subject' sent to ".join(",",@to),4);
  }else{
    pflogger("can not connect to SMTP server $smtpserver!",1);
  }
}

sub isenabled {
  my($enabled) = @_;
  if ($enabled =~ /^\s*(y|yes|true|enable|enabled)\s*$/i) {
    return(1);
  } else {
    return(0);
  }
}

sub isdisabled {
  my($disabled) = @_;
  if ($disabled =~ /^\s*(n|no|false|disable|disabled)\s*$/i) {
    return(1);
  } else {
    return(0);
  }
}

sub getlocalmac {
  my ($dev) = @_;
  return(0) if (!$dev);
  foreach (`/sbin/ifconfig -a`){
    return(clean_mac($1)) if (/^$dev.+HWaddr\s+(\w\w:\w\w:\w\w:\w\w:\w\w:\w\w)/i);
  }
  return(0);
}

sub ip2int {
  return(unpack("N",pack("C4",split(/\./,shift))));
}

sub int2ip {
  return(join(".",unpack("C4",pack("N",shift))));
}

sub get_all_internal_ips {
  my @ips;
  foreach my $interface (@internal_nets) {
    my @tmpips = $interface->enumerate();
    pop @tmpips;
    push @ips, @tmpips;
  }
  return(@ips);
}

sub get_internal_nets {
  my @nets;
  foreach my $interface (@internal_nets) {
    push @nets, $interface->desc();
  }
  return(@nets);
}

sub get_internal_ips {
  my @ips;
  foreach my $internal (@internal_nets) {
    push @ips, $internal->tag("ip");
  }
  return(@ips);
}

sub get_internal_devs {
  my @devs;
  foreach my $internal (@internal_nets) {
    push @devs, $internal->tag("int");
  }
  return(@devs);
}

sub get_internal_devs_phy {
  my @devs;
  foreach my $internal (@internal_nets) {
    my $dev = $internal->tag("int");
    push (@devs, $dev) if ($dev !~ /:\d+$/);
  }
  return(@devs);
}

sub get_external_devs {
  my @devs;
  foreach my $interface (@external_nets) {
    push @devs, $interface->tag("int");
  }
  return(@devs);
}

sub get_managed_devs {
  my @devs;
  foreach my $interface (@managed_nets) {
    push @devs, $interface->tag("int");
  }
  return(@devs);
}

sub get_internal_macs {
  my @macs;
  my %seen;
  foreach my $internal (@internal_nets) {
   my $mac = getlocalmac($internal->tag("int"));
   push @macs, $mac if ($mac && !defined($seen{$mac}));
   $seen{$mac} = 1;
  }
  return(@macs);
}

sub get_internal_info {
  my ($device) = @_;
  foreach my $interface (@internal_nets) {
    return($interface) if ($interface->tag("int") eq $device);
  }
}

sub get_gateways {
  my @gateways;
  foreach my $interface (@internal_nets) {
    push @gateways, $interface->tag("gw");
  }
  return(@gateways);
}

sub get_dhcp_devs {
  my %dhcp_devices;
  foreach my $dhcp (tied(%Config)->GroupMembers("dhcp")) {
   $dhcp_devices{$Config{$dhcp}{'device'}}++ if (defined($Config{$dhcp}{'device'}));
  }
  return(keys(%dhcp_devices));
}


# return 0 if no interfaces
# otherwie return the number of interfaces
#
sub num_interfaces {
  return (scalar(tied(%Config)->GroupMembers("interface")));
}

sub createpid {
  my $pname = basename($0);
  my $pid = $$;
  my $pidfile = $install_dir."/var/$pname.pid";
  pflogger("$pname starting and writing $pid to $pidfile");
  my $outfile = new FileHandle ">$pidfile";
  if (defined($outfile)) {
     print $outfile $pid;
     $outfile->close;
     return($pid);
  } else {
     pflogger("$pname: unable to open $pidfile for writing: $!",1);
     return(-1);
  }
}


sub readpid {
  my ($pname) = @_;
  $pname = basename($0) if (!$pname);
  my $pidfile = $install_dir."/var/$pname.pid";
  my $file = new FileHandle "$pidfile";
  if (defined($file)) {
     my $pid = $file->getline() ;
     $file->close;
     return($pid);
  } else {
     pflogger("$pname: unable to open $pidfile for reading: $!",1);
     return(-1);
  }
}


sub deletepid {
   my ($pname) = @_;
   $pname = basename($0) if (!$pname);
   my $pidfile = $install_dir."/var/$pname.pid";
   unlink($pidfile) || return(-1);
   return(1);
}

sub parse_template {
  my ($tags, $template, $destination) = @_;
  my (@parsed);
  open(TEMPLATE, $template) || die "Unable to open template $template: $!\n";  
  while (<TEMPLATE>) {
    study $_;
    foreach my $tag (keys %{$tags}) {
      $_ =~ s/%%$tag%%/$tags->{$tag}/ig;
    }
    push @parsed, $_;
  }
  #close(TEMPLATE);
  if ($destination) {
    open(DESTINATION, ">".$destination) || die "Unable to open template destination $destination: $!\n";
    foreach my $line (@parsed) {
      print DESTINATION $line;
    }
    #close(DESTINATION);
  } else {
    return(@parsed);
  }
}

sub mysql_date {
  return(POSIX::strftime("%Y-%m-%d %H:%M:%S", localtime));
}

sub util_funnyarp {
  my ($srcmac,$srcip,$destmac,$destip,$type);
  # Check for unicast arp packets
  if ($destmac =~ /ff:ff:ff:ff:ff:ff/i && $destmac =~ /00:00:00:00:00:00/i  &&
       $Config{'arp'}{'listendevice'} ne $monitor_int) {
    pflogger("received unicast ARP from $srcmac ($srcip)  - indicative of a MITM attack",1);
  }
  return(0);
}

sub oui_to_vendor {
  my($mac) = @_;
  my $oui_fh;
  $mac =~ s/:/-/g;
  $mac = uc($mac);
  open($oui_fh, $oui_file) || die "Unable to open $oui_file: $!\n";
  while (<$oui_fh>) {
    chomp;
    if ($_ =~ /^[A-F0-9]{2}\-[A-F0-9]{2}\-[A-F0-9]{2}\s+\(hex\)/) {
      my ($prefix, $hex, $vendor) = split(/\s+/, $_, 3);
      return($vendor) if ($mac =~ /^$prefix/);
    }
  }
  close($oui_fh);
}

sub preload_trappable_ip {
  pflogger("preloading trappable_ip hash", 2);
  return(preload_network_range($Config{'trapping'}{'range'}));
}

sub preload_reggable_ip {
  pflogger("preloading reggable_ip hash", 2);
  return(preload_network_range($Config{'registration'}{'range'}));
}

# Generic Preloading Network Range Function
#
sub preload_network_range {
  my ($network_range) = @_;
    my $caller = (caller(1))[3] || basename($0);
    $caller =~ s/^pf::\w+:://;

  #print "caller: network range = $network_range\n";
  my %cache_ip;

  foreach my $gateway (get_gateways()) {
    $cache_ip{$gateway} = 0;
  }
  foreach my $intip (get_internal_ips()) {
    $cache_ip{$intip} = 0;
  }
  foreach my $range (split(/\s*,\s*/, $network_range)) {
    if ($range =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\/\d{1,2}$/) {
      my $block = new Net::Netmask($range);
      if ($block->size()>2){
        $cache_ip{$block->nth(0)} = 0;
        $cache_ip{$block->nth(-1)} = 0;
      }
      foreach my $ip ($block->enumerate()) {
        $cache_ip{$ip} = 1 if (!defined($cache_ip{$ip}));
      }
    } elsif ($range =~ /^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})-(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})$/) {
      my $start = $1;
      my $end   = $2;
      if (!valid_ip($start) || !valid_ip($end)) {
        pflogger("$range not valid range!",1);
      } else {
        for (my $i = ip2int($start); $i <= ip2int($end); $i++) {
          $cache_ip{int2ip($i)} = 1;
        }
      }
    } elsif ($range =~ /^(\d{1,3}\.\d{1,3}\.\d{1,3})\.(\d{1,3})-(\d{1,3})$/) {
      my $net   = $1;
      my $start = $2;
      my $end   = $3;
      if (!valid_ip($net.".".$start) || $end < $start || $end > 255) {
        pflogger("$range not valid range!", 1);
      } else {
        for (my $i = $start; $ i<= $end; $i++) {
          my $ip = $net.".".$i;
          $cache_ip{$ip} = 1 if (!defined($cache_ip{$ip}));
        }
      }
	} elsif ($range =~ /^(?:\d{1,3}\.){3}\d{1,3}$/){
	  $cache_ip{$range} = 1;  
    } else {
      pflogger("$range not valid!",1);
    }
  }
  pflogger(scalar(keys(%cache_ip))." cache_ip entries cached", 2);
  return(%cache_ip);
}

sub preload_is_internal {
  my %is_internal;
  pflogger("preloading is_internal hash", 2);
  foreach my $interface (@internal_nets) {
    foreach my $ip ($interface->enumerate()) {
      $is_internal{$ip} = 1;
    }
  }
  pflogger(scalar(keys(%is_internal))." is_internal entries cached", 2);
  return(%is_internal);
}

sub throw_hissy_fit {
  my($msg,$pri) = @_;
  $pri = "ERROR" if (!$pri);
  pflogger(uc($pri).": $msg",1);
  print STDERR uc($pri).": $msg\n";          
  exit if ($pri eq "ERROR");
}

1
