#
# $Id: web.pm,v 1.3 2005/11/30 21:41:09 kevmcs Exp $
#
# Copyright 2005 David LaPorte <david@davidlaporte.org>
# Copyright 2005 Kevin Amorin <kev@amorin.org>
#
# See the enclosed file COPYING for license information (GPL).
# If you did not receive this file, see
# http://www.fsf.org/licensing/licenses/gpl.html.
#

package pf::web;

use strict;
use warnings;
use Date::Parse;
use File::Basename;
use Sys::Syslog;
use POSIX();

BEGIN {
  use Exporter ();
  our (@ISA, @EXPORT);
  @ISA    = qw(Exporter);
  @EXPORT = qw(generate_release_page generate_error_page generate_status_page generate_registration_page parse_auth_conf);
}

use lib qw(/usr/local/pf/lib);
use pf::config;
use pf::util;
use pf::iplog qw(ip2mac);
use pf::node qw(node_view);
use pf::violation qw(violation_view_open);


sub generate_release_page {
  my ($destination_url,$msg) = @_;
  my %tags;
  $msg = " " if (!$msg);
  my $q = new CGI;
  print $q->header;
  $tags{'logo'} = $Config{'general'}{'logo'};
  $tags{'timer'} = $Config{'trapping'}{'redirtimer'};
  $tags{'destination_url'} = $destination_url;
  $tags{'message'} = $msg;
  print STDOUT parse_template(\%tags, "$install_dir/html/user/content/templates/release.html"); 
  exit;
}

sub generate_error_page {
  my ($error_msg) = @_;
  my %tags;
  my $q = new CGI;
  print $q->header;
  $tags{'logo'} = $Config{'general'}{'logo'};
  $tags{'error_msg'} = $error_msg;
  print STDOUT parse_template(\%tags, "$install_dir/html/user/content/templates/error.html"); 
  exit;
}

# ugly hack - fix me!
sub generate_status_page {
  my ($mac) = @_;
  my %tags;
  my $q = new CGI;
  my $ip = $q->remote_addr;
  print $q->header;

  $tags{'logo'} = $Config{'general'}{'logo'};
  $tags{'node_ip'} =$ip;
  $tags{'node_mac'} = ip2mac($ip);
  $tags{'gateway_ip'} = ip2gateway($ip);
  $tags{'gateway_mac'} = ip2mac(ip2gateway($ip));
  $tags{'pf_mac'} = getlocalmac(ip2device($ip));

  my $node_info = node_view($mac);
  $tags{'pid'} = $node_info->{'pid'};
  $tags{'computername'} = $node_info->{'computername'};

  my $reg_status = $node_info->{'status'};
  $reg_status="Registered"   if ($reg_status =~ /^reg$/i);
  $reg_status="Unregistered" if ($reg_status =~ /^unreg$/i);
  $reg_status="Grace"        if ($reg_status =~ /^grace$/i);
  $tags{'reg_status'} = $reg_status;           

  my @violations = violation_view_open($mac);
  my $violations;

  foreach my $violation (@violations) {
    my $class_info = class_view($violation->{'vid'});
    $violations.="<td>$class_info->{'description'}</td><td>$violation->{'status'}</td>";
  }
  $violations = '<td colspan="2">NONE</td>' if (!$violations);
  $tags{'violations'} = $violations;

  print STDOUT parse_template(\%tags, "$install_dir/html/user/content/templates/status.html"); 
  exit;
}

sub generate_registration_page {
  my ($destination_url,$mac) = @_;
  my %tags;
  my $q = new CGI;
  print $q->header;
  $tags{'logo'} = $Config{'general'}{'logo'};
  $tags{'deadline'} = $Config{'registration'}{'skip_deadline'};
  $tags{'hostname'} = $Config{'general'}{'hostname'};
  $tags{'domain'} = $Config{'general'}{'domain'};
  $tags{'destination_url'} = $destination_url;
  $tags{'button_text'} = $Config{'registration'}{'button_text'};

  # check to see if node can skip reg 
  my $skip = 0;
  my $node_info = node_view($mac);
  my $detect_date = str2time($node_info->{'detect_date'});
  my $registration_mode = $Config{'registration'}{'skip_mode'};
  
  my $skip_allowed_until = 0;
  if ($registration_mode eq "deadline") {
    $skip_allowed_until = $Config{'registration'}{'deadline'};
  } elsif ($registration_mode eq "window") {
    $skip_allowed_until = $detect_date + $Config{'registration'}{'window'};
  }
  
  my $skip_until = POSIX::strftime("%Y-%m-%d %H:%M:%S", POSIX::localtime($skip_allowed_until));
  if (time < $skip_allowed_until) {
    pflogger($node_info->{'mac'} . " allowed to skip registration until $skip_until", 8);
    $skip = 1;
  } else {
    pflogger($node_info->{'mac'} . " is not allowed to skip registration - deadline passed at $skip_until - ", 8);
  }
  
  my @pulldown;
  my @auth=split(/\s*,\s*/,$Config{'registration'}{'auth'});
  #
  # if no skip and one Auth type you don't need a pull down...
  if (scalar(@auth)==1 && !$skip) {
    push @pulldown,"<input type='hidden' name='auth' value='$auth[0]'>";
  } else {
    push @pulldown,"<select name='auth'>";
    push @pulldown,"<option value='' selected> Select Authentication";
    foreach my $auth (@auth){
	  push @pulldown,"<option value='$auth'> ".parse_auth_conf($auth);
    }
	push @pulldown,"<option value='skip'> Skip Registration" if ($skip);
    push @pulldown,"</select>";
  }
  
  $tags{'pulldown'}=join("\n",@pulldown);
  print STDOUT parse_template(\%tags, "$install_dir/html/user/content/templates/register.html"); 
  exit;
}

sub parse_auth_conf {
  my ($authtype)=@_;
  my $authconf = "$install_dir/conf/templates/${authtype}.conf";
  if (-e $authconf) {
     open(AUTHCONF, $authconf) || die("Unable to open authentication config file $authconf: $!\n",1); 
     while (<AUTHCONF>){
	   return $1 if (/AuthName\s+"*(.+?)"*\s*$/i);
     }
  }else{
     pflogger("Auth File $authconf does not exist",1);
  }
  return $authtype;
} 

1
